# Python script
# by Jakub Imriska, ITP, ETH Zurich, 2013-2014
# for course on Mechanics of Continua by Vadim Geshkenbein (FS 2014)
# Homework 10.4
# Task: find the stationary 2-dimensional flow of viscous incompressible fluid around a rectangular barrier

# CORE FUNCTIONS

# libraries for (sparse) linear algebra
import numpy as np
import scipy, scipy.sparse
from scipy.sparse.linalg import spsolve

# import input parameters
from parameters import *


def initialize():
  # initial guess for vorticity and stream function according to the free flow far from the barrier
  # note that in the program we use the dimensionsless (rescaled) vorticity and stream function
  #   the stream function is in units of (v_0*h)
  #   the vorticity is in units of (v_0/h)
  
  # initial guess for the stream function
  stream_func = np.zeros((Nx,Ny),np.float64)
  for x in range(0,Nx):
    y_low = 1
    if (not debug_mode) and (x>=a and x<=a+t):  # range of the barrier (including its boundaries)
      y_low = w+1
    for y in range(y_low,Ny):  # away from the barrier and its boundaries
      stream_func[x,y] = y * (1. - ((y/(Ny-1.))**2)/3.)  # stream function of the flow far from the barrier
  
  # initial guess for the vorticity
  vorticity = np.zeros((Nx,Ny),np.float64)
  for x in range(0,Nx):
    y_low = 1
    if (not debug_mode) and (x>a and x<a+t):  # range of the barrier (excluding its boundaries)
      y_low = w
    for y in range(y_low,Ny):  # away from the interior of the barrier
      vorticity[x,y] = -2.*y/(Ny-1.)**2   # vorticity of the flow far from the barrier
  
  return (stream_func,vorticity)


# multiindex mapping of the pair of coordinates x and y
def index(x,y):
  return x+y*Nx


def solve_stream_func(stream_func,vorticity,relaxation_rate):
  # this procedure does find the new stream_func according to the vorticity and boundary conditions
  
  if not hasattr(solve_stream_func, "A_csc"):
    # we need to initialize the matrix A in the 1st function call
    A = scipy.sparse.dok_matrix((Nx*Ny,Nx*Ny),np.float64)  # we build the matrix using dictionary format
    for x in range(0,Nx):
      for y in range(0,Ny):
        i = index(x,y)
        if x==0:  # left boundary: Neumann condition: \partial_x stream_func = 0
          A[i,i] = 1
          A[i,index(1,y)] = -1
        elif x==Nx-1:  # right boundary: Neumann condition: \partial_x stream_func = 0
          A[i,i] = 1
          A[i,index(Nx-2,y)] = -1
        elif y==0 or y==Ny-1:  # top boundary and the symmetry axis: Dirichlet condition
          A[i,i] = 1
        elif (not debug_mode) and (y<=w and x>=a and x<=a+t):  # in the barrier (including boundaries): Dirichlet condition
          A[i,i] = 1
        else:  # away from boundaries and barrier: discretized partial differential equation
          A[i,index(x+1,y)] = 1
          A[i,index(x-1,y)] = 1
          A[i,index(x,y+1)] = 1
          A[i,index(x,y-1)] = 1
          A[i,i] = -4
    solve_stream_func.A_csc = scipy.sparse.csc_matrix(A)  # we convert the sparse matrix into Compressed Sparse Column format
  
  q = np.zeros(Nx*Ny,np.float64)
  for x in range(0,Nx):
    for y in range(0,Ny):
      if x==0 or y==0 or x==Nx-1:  # left (Neumann) and right (Neumann) boundaries and symmetry axis (Dirichlet)
        continue
      elif y==Ny-1:  # top boundary (Dirichlet)
        q[index(x,y)] = stream_func[x,y]
      elif (not debug_mode) and (y<=w and x>=a and x<=a+t):  # in the barrier (including boundaries) (Dirichlet)
        continue  # Alternative: q[index(x,y)]=0.  NOTE the stream function has to be zero there; as it is initialized with zeros, there is no need of any action
      else:  # away from boundaries and barrier
        q[index(x,y)] = vorticity[x,y]
  
  stream_func_new = scipy.sparse.linalg.spsolve(solve_stream_func.A_csc,q)
  
  max_diff = 0.
  for x in range(0,Nx):
    for y in range(0,Ny):
      i = index(x,y)
      if abs(stream_func[x,y]-stream_func_new[i])>max_diff:
        max_diff = abs(stream_func[x,y]-stream_func_new[i])
      stream_func[x,y] = (1.-relaxation_rate)*stream_func[x,y] + relaxation_rate*stream_func_new[i]  # mixing the old and new solution
  
  return (stream_func,max_diff)


def solve_vorticity(stream_func,vorticity,relaxation_rate):
  # this procedure does find the new vorticity according to the stream_func and boundary conditions
  
  B = scipy.sparse.dok_matrix((Nx*Ny,Nx*Ny),np.float64)  # sparse matrix in dictionary format
  r = np.zeros(Nx*Ny,np.float64)
  for x in range(0,Nx):
    for y in range(0,Ny):
      i = index(x,y)
      if x==0:  # the left boundary: Neumann condition: \partial_x vorticity = 0
        B[i,i] = 1
        B[i,index(x+1,y)] = -1
      elif x==Nx-1:  # the right boundary: Neumann condition: \partial_x vorticity = 0
        B[i,i] = 1
        B[i,index(x-1,y)] = -1
      elif y==0:  # symmetry axis: Dirichlet condition
        B[i,i] = 1
      elif y==Ny-1:  # the top boundary: Dirichlet condition
        B[i,i] = 1
        r[i] = 2*(stream_func[x,y-1]-stream_func[x,y])
      elif (not debug_mode) and (y<=w and x>=a and x<=a+t):  # within the barrier or on its boundaries
        if (y<w and x>a and x<a+t):  # inside of the barrier (excluding its boundaries): Dirichlet condition for dummy variables
          B[i,i] = 1  # has to be non-zero      
        elif x==a and y<w:  # the left boundary of the barrier (excluding its top left corner): Dirichlet condition
          B[i,i] = 1
          r[i] = 2*stream_func[x-1,y]
        elif x==a+t and y<w:  # the right boundary of the barrier (excluding its top right corner): Dirichlet condition
          B[i,i] = 1
          r[i] = 2*stream_func[x+1,y]
        elif x>a and x<a+t and y==w:  # the top boundary of the barrier (excluding both corners): Dirichlet condition
          B[i,i] = 1
          r[i] = 2*stream_func[x,y+1]
        elif x==a and y==w:  # the top left corner of the barrier: Dirichlet condition
          B[i,i] = 1
          r[i] = stream_func[x-1,y] + stream_func[x,y+1]
        elif x==a+t and y==w:  # the top right corner of the barrier: Dirichlet condition
          B[i,i] = 1
          r[i] = stream_func[x+1,y] + stream_func[x,y+1]
      else: # away from boundaries and barrier
        B[i,index(x+1,y)] = 1 - 0.25*lattice_Re*(stream_func[x,y+1]-stream_func[x,y-1])
        B[i,index(x-1,y)] = 1 + 0.25*lattice_Re*(stream_func[x,y+1]-stream_func[x,y-1])
        B[i,index(x,y+1)] = 1 + 0.25*lattice_Re*(stream_func[x+1,y]-stream_func[x-1,y])
        B[i,index(x,y-1)] = 1 - 0.25*lattice_Re*(stream_func[x+1,y]-stream_func[x-1,y])
        B[i,i] = -4
  
  vorticity_new = spsolve(scipy.sparse.csc_matrix(B),r)
  
  max_diff = 0.
  for x in range(0,Nx):
    for y in range(0,Ny):
      i = index(x,y)
      if abs(vorticity[x,y]-vorticity_new[i])>max_diff:
        max_diff = abs(vorticity[x,y]-vorticity_new[i])
      vorticity[x,y] = (1.-relaxation_rate)*vorticity[x,y] + relaxation_rate*vorticity_new[i]  # mixing the old and new solution
  
  return (vorticity,max_diff)
