# Python script
# by Jakub Imriska, ITP, ETH Zurich, 2013-2014
# for course on Mechanics of Continua by Vadim Geshkenbein (FS 2014)
# Homework 10.4
# Task: find the stationary 2-dimensional flow of viscous incompressible fluid around a rectangular barrier

# 2D PLOTTING OF THE RESULTING STREAMFUNCTION AND VORTICITY

# library for plotting
import matplotlib
import matplotlib.pyplot as plt
import numpy as np

# import input parameters
from parameters import *

# switch to plot the upper half only, otherwise the full problem will be plotted (using the symmetry)
show_upper_half_only = True


def colormap_pos_neg(min_,max_):
  # this function returns colormap for the range [min_,max_]
  # positive values are given in blue tones [larger values are more saturated]
  # negative values (if there are any) are given in red tones [larger magnitudes are more saturated]
  if min_>=0:
    cdict = {'red':   [(0.0,  1.0, 1.0),
                       (1.0,  0.0, 1.0)],
             'green': [(0.0,  1.0, 1.0),
                       (1.0,  0.0, 1.0)],
             'blue':  [(0.0,  1.0, 1.0),
                       (1.0,  1.0, 1.0)]}
  else:
    if -10.*min_<max_:
      d = -10.*min_/max_
    else:
      d = 1.
    zero_ = -min_/(max_-min_)
    cdict = {'red':   [(0.0,  0.0, 1.0),
                       (zero_,  1.0, 1.0),
                       (1.0,  0.0, 1.0)],
             'green': [(0.0,  0.0, 1.-d),
                       (zero_, 1.0, 1.0),
                       (1.0,  0.0, 1.0)],
             'blue':  [(0.0,  0.0, 1.-d),
                       (zero_,  1.0, 1.0),
                       (1.0,  1.0, 1.0)]}
  return matplotlib.colors.LinearSegmentedColormap('pos_neg',cdict)


def graymap_pos_neg(min_,max_):
  # gray variant of the colormap_pos_neg
  # this function returns colormap for the range [min_,max_]
  # positive values are given in gray tones [larger values are darker]
  # negative values (if there are any) are given in gray tones [larger magnitudes are darker]
  if min_>=0:
    cdict = {'red':   [(0.0,  1.0, 1.0),
                       (1.0,  0.333, 1.0)],
             'green': [(0.0,  1.0, 1.0),
                       (1.0,  0.333, 1.0)],
             'blue':  [(0.0,  1.0, 1.0),
                       (1.0,  0.333, 1.0)]}
  else:
    if -10.*min_<max_:
      d = -10.*min_/max_
    else:
      d = 1.
    zero_ = -min_/(max_-min_)
    cdict = {'red':   [(0.0,  0.0, 1.0-2*d/3),
                       (zero_,  1.0, 1.0),
                       (1.0,  0.333, 1.0)],
             'green': [(0.0,  0.0, 1.-2*d/3),
                       (zero_, 1.0, 1.0),
                       (1.0,  0.333, 1.0)],
             'blue':  [(0.0,  0.0, 1.-2*d/3),
                       (zero_,  1.0, 1.0),
                       (1.0,  0.333, 1.0)]}
  return matplotlib.colors.LinearSegmentedColormap('pos_neg',cdict)


def show(stream_func,vorticity):
  # function to show plot with vorticity and plot with stream function
  max_stream_func = stream_func.max()
  min_stream_func = stream_func.min()
  h = 1./points_per_unit_length
  x = np.arange(0.0, L+0.5*h, h)
  y = np.arange(0.0, H+0.5*h, h)
  X, Y = np.meshgrid(x, y)  # mesh for the upper half
  y_lower = np.arange(0.0, -H-0.5*h, -h)
  X_lower, Y_lower = np.meshgrid(x, y_lower)  # symmetrically mirrored mesh for the lower half
  
  # stream function plot
  plt.figure(figsize=(12,8))
  plt.title('Stream function and streamlines for $Re={:.4}$ ($h={:.4}$)'.format(physical_Re,h))
  ax = plt.axes()
  ax.set_xlim(0.0,L)
  im = plt.imshow(stream_func.T, interpolation='bilinear', origin='lower', cmap=colormap_pos_neg(min_stream_func,max_stream_func), extent=(0,L,0,H))  # plot of the stream function
  im_lower = plt.imshow(stream_func.T, interpolation='bilinear', origin='upper', cmap=graymap_pos_neg(min_stream_func,max_stream_func), extent=(0,L,-H,0))  # plot of the stream function in the lower half
  CBI = plt.colorbar(im, orientation='horizontal')  # legend with color code
  CBI.set_label(r'dimensionless stream function (in units $v_0 h$)')
  levels = np.arange(max_stream_func/16.,max_stream_func,max_stream_func/16.).tolist()  # streamlines of normal flow
  if min_stream_func<-0.001:  # is there a significant eddy current? if yes, then we plot the corresponding streamlines
    CS0 = plt.contour(X, Y, stream_func.T, [0.], colors='k', linestyles='dotted', linewidths=2)  # contour separating normal flow and eddy currents
    CS0_lower = plt.contour(X_lower, Y_lower, stream_func.T, [0.], colors='k', linestyles='dotted', linewidths=2)  # contour separating normal flow and eddy currents
    for i in range(1,4):  # streamlines of eddy current
      levels.append(i*min_stream_func/4.)
  CS = plt.contour(X, Y, stream_func.T, levels, colors='k')  # plotting the streamlines
  CS_lower = plt.contour(X_lower, Y_lower, stream_func.T, levels, colors='k')  # plotting the streamlines in the lower half 
  plt.gca().add_patch(plt.Rectangle((0.0, H), L, 0.2*H ,facecolor="#DDDDDD",hatch='/',zorder=100000))  # the upper boundary
  plt.gca().add_patch(plt.Rectangle((0.0, -1.2*H), L, 0.2*H ,facecolor="#DDDDDD",hatch='/',zorder=100000))  # the lower boundary
  plt.axhline(linestyle='-.',color='k')      # symmetry axis (which itself is a streamline)
  if show_upper_half_only:
    ax.set_ylim(0,H)
  else:
    ax.set_ylim(-1.1*H,1.1*H)
  if not debug_mode:
    plt.gca().add_patch(plt.Rectangle((F, -W), T, 2*W, facecolor="#808080",zorder=100000))  # plotting the barrier
  
  # vorticity plot
  plt.figure(figsize=(12,8))
  plt.title('Vorticity for $Re={:.4}$ ($h={:.4}$)'.format(physical_Re,h))
  ax = plt.axes()
  ax.set_xlim(0.0,L)
  im = plt.imshow(vorticity.T, interpolation='bilinear', origin='lower', cmap=colormap_pos_neg(vorticity.min(),vorticity.max()), extent=(0,L,0,H))  # plot of the vorticity
  im_lower = plt.imshow(vorticity.T, interpolation='bilinear', origin='upper', cmap=graymap_pos_neg(vorticity.min(),vorticity.max()), extent=(0,L,-H,0))  # plot of the vorticity in the lower half
  CBI = plt.colorbar(im, orientation='horizontal')  # legend with color code
  CBI.set_label(r'dimensionless vorticity (in units $v_0/ h$)')
  CS = plt.contour(X, Y, vorticity.T, [0.], colors='k', linestyles='dotted', linewidths=2)  # contour between the regions with positive and negative vorticity
  CS_lower = plt.contour(X_lower, Y_lower, vorticity.T, [0.], colors='k', linestyles='dotted', linewidths=2)  # contour between the regions with positive and negative vorticity
  plt.gca().add_patch(plt.Rectangle((0.0, H), L, 0.2*H ,facecolor="#DDDDDD",hatch='/',zorder=100000))  # the upper boundary
  plt.gca().add_patch(plt.Rectangle((0.0, -1.2*H), L, 0.2*H ,facecolor="#DDDDDD",hatch='/',zorder=100000))  # the lower boundary
  plt.axhline(linestyle='-.',color='k')  # symmetry axis (which separates the positive and negative vorticity)
  if show_upper_half_only:
    ax.set_ylim(0,H)
  else:
    ax.set_ylim(-1.1*H,1.1*H)
  if not debug_mode:
    plt.gca().add_patch(plt.Rectangle((F, -W), T, 2*W, facecolor="#808080",zorder=100000))  # plotting the barrier

  plt.show()
